'' =================================================================================================
''
''   File....... jm_flip_holiday_kit_tester.spin
''   Purpose.... Test IO devices for FLiP Holiday Kit
''   Author..... Jon McPhalen
''               Copyright (c) 2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started.... 21 DEC 2024
''   Updated.... 22 DEC 2024
''
'' =================================================================================================


con { timing }

  _xinfreq = 5_000_000                                          ' use 5MHz crystal
  _clkmode = xtal1 + pll16x                                     ' x16 = 80MHz

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  PGM_RX   = 31  { I }                                          ' serial / programming
  PGM_TX   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }

  LED2     = 27  { O }                                          ' FLiP LEDs
  LED1     = 26  { O }


con { app io pins }

  OLED_CS  = 23  { O }                                          ' OLED connections
  OLED_DC  = 22  { O }
  OLED_RST = 21  { O }
  OLED_DAT = 20  { O }
  OLED_CLK = 19  { O }
  OLED_3V3 = 18  { P }
  OLED_GND = 17  { P }

  PIX_OUT  = 16  { O }                                          ' WS2812b pixels

  JOY2_3V3 = 15  { P }                                          ' right stick
  JOY2_CB  = 14  { I }
  JOY2_LF  = 13  { I }
  JOY2_GND = 12  { P }
  JOY2_DN  = 11  { I }
  JOY2_RT  = 10  { I }
  JOY2_UP  =  9  { I }

  PIEZO    =  8  { O }                                          ' for beeps, boops, music

  JOY1_DN  =  6  { I }                                          ' left stick
  JOY1_RT  =  5  { I }
  JOY1_UP  =  4  { I }
  JOY1_3V3 =  3  { P }
  JOY1_CB  =  2  { I }
  JOY1_LF  =  1  { I }
  JOY1_GND =  0  { P }


con

  #true,  ON, OFF
  #false, NO, YES

  USE_TERMINAL = YES


con

  N_PIX  = 50

  RED    = $10_00_00_00                                         ' primaries
  GREEN  = $00_10_00_00
  BLUE   = $00_00_10_00

  ORANGE = $0E_02_00_00                                         ' mixed
  YELLOW = $0B_05_00_00
  CYAN   = $00_08_08_00
  VIOLET = $08_00_08_00
  WHITE  = $05_05_05_00

  BLACK  = $00_00_00_00


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  prng : "jm_prng"                                              '   pseudo-random number generation
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  oled : "jm_ssd1306_128x64_ez"                                 '   OLED driver
  rgb  : "jm_ws2811"                                            ' * WS2811 pixel driver

' * uses cog when loaded


dat { pre-initialized }

  Menu          byte    "=======================", 13
                byte    "FLiP Holiday Kit Tester", 13
                byte    "=======================", 13
                byte    13
                byte    " [1] OLED              ", 13
                byte    " [2] Pixels            ", 13
                byte    " [3] 5-way Buttons     ", 13
                byte    " [4] Piezo             ", 13
                byte    0


var { globals }

  long  pixbuf[N_PIX]
  long  pixpad[4]                                               ' pad for aninmation fills


pub main | k

  setup
  if (USE_TERMINAL == YES)
'   wait_for_terminal(true, 250)

  repeat
    term.tx(term#CLS)
    term.str(@Menu)
    term.rxflush
    k := term.rx

    case k
      "1" : test_oled
      "2" : test_pixels
      "3" : test_buttons
      "4" : test_piezo

  repeat
    waitcnt(0)


pub test_oled

  oled.clear
  time.pause(100)

  oled.lstr(string("PARALLAX"), oled#FONT_NORM, oled#PIXEL_ON, 0)
  oled.box_fill(30, 20, 95, 41, oled#PIXEL_ON)
  oled.cstr(string("FLiP"), oled#FONT_BIG, oled#PIXEL_OFF, 3)
  oled.rstr(string("HOLIDAY KIT"), oled#FONT_NORM, oled#PIXEL_ON, 7)


dat

  PixMenu       byte    "Test Pixels       ", 13
                byte    "----------------- ", 13
                byte    13
                byte    " [R] Red          ", 13
                byte    " [G] Green        ", 13
                byte    " [B] Blue         ", 13
                byte    " [F] Flash        ", 13
                byte    " [M] Marquee      ", 13                    ' last color
                byte    " [C] Christmas    ", 13                    ' red/white/green
                byte    " [?] Randomize    ", 13
                byte    " [!] Roll Rainbow ", 13
                byte    " [Q] Quit         ", 13
                byte    0

  PressKey      byte    "Press any key to quit", 0


pub test_pixels : last | k

  repeat
    term.tx(term#CLS)
    term.str(@PixMenu)
    term.rxflush
    k := term.rx
    case k
      "r", "R" : last := show_color(RED)
      "g", "G" : last := show_color(GREEN)
      "b", "B" : last := show_color(BLUE)
      "f", "F" : run_flash(last)
      "m", "M" : run_marquee(last)
      "c", "C" : run_christmas
      "?"      : run_random
      "!"      : run_rainbow
      "q", "Q" : return


pri show_color(c) : result | i

  result := c

  term.tx(term#CLS)
  term.fstr1(string("Color: $%.8x\r"), c)

  rgb.clear

  repeat i from 0 to N_PIX-1
    rgb.set(i, c)
    time.pause(25)

  time.pause(1000)
  rgb.clear


pri run_flash(c)

  if (c == BLACK)
    c := RED

  term.tx(term#CLS)
  term.fstr1(string("Flashing with: $%.8x\r"), c)
  term.str(@PressKey)

  term.rxflush

  repeat
    rgb.set_all(c)
    repeat 5
      if (term.rxtime(100) => 0)
        rgb.clear
        return

    rgb.set_all(BLACK)
    repeat 5
      if (term.rxtime(100) => 0)
        rgb.clear
        return


pri run_marquee(c) | i, j, nb, block[5]

  if (c == BLACK)
    c := RED

  nb := N_PIX / 3 + 1                                           ' # of 3-pixel blocks

  longfill(@block, BLACK, 5)
  block[2] := c

  term.tx(term#CLS)
  term.fstr1(string("Marquee with: $%.8x\r"), c)
  term.str(@PressKey)

  term.rxflush

  repeat
    repeat i from 2 to 0                                        ' 3-step sequence
      j := 0                                                    ' start at head of string
      repeat nb                                                 ' repeat for each 3-pixel block
        longmove(@pixbuf[j], @block[i], 3)                      ' update block
        j += 3                                                  ' point to next block
      if (term.rxtime(150) => 0)                                ' hold / check key input
        rgb.clear
        return


dat

  RWGx          long    WHITE, GREEN, BLACK, RED, WHITE, GREEN, BLACK


pri run_christmas | nb, i, j

' Slower marquee with Christmas colors

  term.tx(term#CLS)
  term.fstr0(string("Christmas colors marquee\r"))
  term.str(@PressKey)

  term.rxflush

  nb := N_PIX >> 2 + 1                                          ' # of 4-pixel blocks

  repeat
    repeat i from 3 to 0                                        ' 4-step sequence
      j := 0                                                    ' start at head of string
      repeat nb                                                 ' repeat for each 4-pixel block
        longmove(@pixbuf[j], @RWGx[i], 4)                       ' update RWGx block
        j += 4                                                  ' point to next block
      if (term.rxtime(250) => 0)                                ' hold / check key input
        rgb.clear
        return


dat

  Colors        long    RED, ORANGE, YELLOW, GREEN, CYAN, BLUE, VIOLET, WHITE


pri run_random | i, j

  term.tx(term#CLS)
  term.fstr0(string("Random colors\r"))
  term.str(@PressKey)

  term.rxflush

  repeat
    repeat i from 0 to constant(N_PIX-1)
      j := prng.randomize(0, 7)
      rgb.set(i, Colors[j])
    if (term.rxtime(1000) => 0)
      rgb.clear
      return


pri run_rainbow : phase | i

  term.tx(term#CLS)
  term.fstr0(string("Rolling rainbow\r"))
  term.str(@PressKey)

  term.rxflush

  repeat
    repeat i from 0 to constant(N_PIX-1)
      rgb.set(i, rgb.wheelx((i + phase) * constant(255/N_PIX), $10))
    if (term.rxtime(25) => 0)
      rgb.clear
      return
    --phase                                                     ' use ++ for reverse


dat

  BtnMenu       byte    "Test Buttons   ", 13
                byte    "-------------- ", 13
                byte    13
                byte    " ( ) 1 Up      ", 13                   ' @ 2,3
                byte    " ( ) 1 Down    ", 13
                byte    " ( ) 1 Left    ", 13
                byte    " ( ) 1 Right   ", 13
                byte    " ( ) 1 Center  ", 13
                byte    13
                byte    " ( ) 2 Up      ", 13                   ' @ 2, 9
                byte    " ( ) 2 Down    ", 13
                byte    " ( ) 2 Left    ", 13
                byte    " ( ) 2 Right   ", 13
                byte    " ( ) 2 Center  ", 13
                byte    13
                byte    0


pub test_buttons

  term.tx(term#CLS)
  term.str(@BtnMenu)
  term.str(@PressKey)
  term.rxflush

  repeat
    term.fstr1(string("\002\002\003%c"), 42-(ina[JOY1_UP]*10))
    term.fstr1(string("\002\002\004%c"), 42-(ina[JOY1_DN]*10))
    term.fstr1(string("\002\002\005%c"), 42-(ina[JOY1_LF]*10))
    term.fstr1(string("\002\002\006%c"), 42-(ina[JOY1_RT]*10))
    term.fstr1(string("\002\002\007%c"), 42-(ina[JOY1_CB]*10))

    term.fstr1(string("\002\002\009%c"), 42-(ina[JOY2_UP]*10))
    term.fstr1(string("\002\002\010%c"), 42-(ina[JOY2_DN]*10))
    term.fstr1(string("\002\002\011%c"), 42-(ina[JOY2_LF]*10))
    term.fstr1(string("\002\002\012%c"), 42-(ina[JOY2_RT]*10))
    term.fstr1(string("\002\002\013%c"), 42-(ina[JOY2_CB]*10))

    if (term.rxtime(25) => 0)
      return


dat

  SndMenu       byte    "Tones & Music           ", 13
                byte    "----------------------- ", 13
                byte    13
                byte    " [W] Welcome            ", 13
                byte    " [!] Warn               ", 13
                byte    " [S] Success            ", 13
                byte    " [F] Failure            ", 13
                byte    " [H] Hello              ", 13
                byte    " [C] Charge!            ", 13
                byte    " [T] Oh, Christmas Tree ", 13
                byte    " [D] Dreidel Song       ", 13
                byte    " [Q] Quit               ", 13
                byte    0


dat { beeps, boops, music }

' MM

  Welcome       long    248_9, 100, 234_9, 100, 221_7, 100, 203_9, 100
                long    0, 0, -1, -1

  Warn          long    N_F3, T_32
                long    0, 0, -1, -1

  Success       long    N_A3, T_32, REST, T_64, N_A3, T_32
                long    0, 0, -1, -1

  Failure       long    N_F2, T_32
                long    0, 0, -1, -1

' JM

  Hello         long    N_E6, T_8, N_C5, T_8, N_G5, T_4
                long    0, 0, -1, -1

  Charge        long    N_G4, T_8,  N_C5, T_8, N_E5, T_8
                long    N_G5, T_D8, N_E5, T_8, N_G5, T_2
                long    0, 0, -1, -1

  Christmas     long    N_C4, T_4, N_F4, T_D8, REST, T_X, N_F4, T_16, REST,  T_X, N_F4, T_4
                long    N_G4, T_4, N_A4, T_D8, REST, T_X, N_A4, T_16, REST,  T_X, N_A4, T_D4, REST, T_X
                long    N_A4, T_8,  N_G4, T_8, N_A4, T_8,  N_As4, T_4, N_E4, T_4, N_G4, T_4, N_F4, T_4
                long    0, 0, -1, -1

  Dreidel       long    N_G4, T_4, N_C4, T_4, REST, T_X, N_C4, T_4, N_D4, T_4, REST, T_X, N_D4, T_4, N_E4, T_4, N_C4, T_2
                long    N_E4, T_4, N_G4, T_4, REST, T_X, N_G4, T_4, REST, T_X, N_F4, T_4, N_E4, T_4, N_D4, T_D2
                long    N_D4, T_4, REST, T_X, N_D4, T_4, REST, T_X, N_D4, T_4, REST, T_X,  N_E4, T_4, REST, T_X, N_E4, T_4, REST, T_X, N_F4, T_4, N_D4, T_2
                long    N_D4, T_4, N_G4, T_4, N_F4, T_4, N_E4, T_4, N_D4, T_4, N_C4, T_D2
                long    0, 0, -1, -1


pub test_piezo | k

  term.tx(term#CLS)
  term.str(@SndMenu)

  term.rxflush

  repeat
    k := term.rxtime(25)
    case k
      "w", "W" : start_music(@Welcome,    60)
      "!"      : start_music(@Warn   ,   120)
      "s", "S" : start_music(@Success,   120)
      "f", "F" : start_music(@Failure,   120)
      "h", "H" : start_music(@Hello,     240)
      "c", "C" : start_music(@Charge,    180)
      "t", "T" : start_music(@Christmas, 120)
      "d", "D" : start_music(@Dreidel,   180)
      "q", "Q" : quit

  stop_music


pub setup

'' Setup IO and objects for application

  time.start                                                    ' setup timing & delays

  prng.seed(-cnt, -cnt ~> 2, $CAFE_BABE, cnt <- 2, cnt)         ' seed randomizer

  io.start(0, 0)                                                ' clear all pins (master cog)
  setup_buttons

  setup_oled

  rgb.start(@pixbuf, N_PIX, PIX_OUT)                            ' start ws2811 driver *

  bcog := cognew(background, @bcstack) + 1                      ' start background cog *

  if (USE_TERMINAL == YES)
    term.tstart(BR_TERM)                                        ' start serial for terminal *


pri setup_buttons

'' Configure IO for 5-way switches
'' -- power & ground provided by FLiP IO pins

  io.low(JOY1_GND)                                              ' enable ground
  io.high(JOY1_3V3)                                             ' activate pull-ups

  io.low(JOY2_GND)
  io.high(JOY2_3V3)


pri setup_oled

'' Configure and initialize OLED
'' -- power & ground provided by FLiP IO pins

  io.low(OLED_GND)                                              ' power up oled
  time.pause(5)
  io.high(OLED_3V3)
  time.pause(20)

  oled.startx(OLED_RST, OLED_DC, OLED_CS, OLED_DAT, OLED_CLK)   ' initialize oled
  oled.flip(true)                                               ' flip screen for holiday kit


pub wait_for_terminal(clear, delay)

'' Wait for terminal to be open and key pressed

  if (USE_TERMINAL <> YES)
    return

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)
    time.pause(delay #> 0)


con { bkg support }

  { ----------------------------------------- }
  {  B / G   S U P P O R T   R O U T I N E S  }
  { ----------------------------------------- }


pub start_tone(freq, ms)

'' Play tone of freq for ms milliseconds
'' -- non-blocking
'' -- freq in 0.1Hz units

  if (p_music > 0)                                              ' busy
    longfill(@p_music, 0, 2)                                    ' kill music
    waitcnt(cnt + MS_001)                                       ' - give it time to die

  longmove(@OneNote, @freq, 2)                                  ' setup note

  BPM := 60                                                     ' play it
  p_music := @OneNote


pub stop_tone

'' Kills background note if playing

  longfill(@p_music, 0, 2)


pub start_music(p_notes, rate)

'' Play music (notes/timing) at p_notes
'' -- non-blocking
'' -- rate is BPM

  BPM     := rate
  p_music := p_notes


pub stop_music

'' Kills background music if playing

  longfill(@p_music, 0, 2)


pub playing

'' Returns true when background tone/music is playing

  return (p_music > 0)


con { background }

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - play monophonic music      }
  {  - play simple tone           }
  { ----------------------------- }


var

  long  bcog                                                    ' background cog #
  long  bcstack[64]                                             ' stack for background cog

  long  millis                                                  ' global milliseconds register


pri background | t                                              ' launch with cognew()

  millis := 0

  t := cnt                                                      ' sync loop timer
  repeat
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis
    play_music


var

  long  p_music
  long  nmillis


dat

  BPM           long    120

  OneNote       long    0, 0, -1, -1                            ' for tone_on


pri play_music | nfreq

  if (p_music =< 0)                                             ' manual stop invoked?
    io.freq_out(0, PIEZO, 0)
    return

  if (--nmillis > 0)                                            ' is note playing now?
    return

  nfreq := long[p_music]                                        ' get note
  p_music += 4                                                  ' advance to timing
  if (nfreq < 0)                                                ' if end
    longfill(@p_music, 0, 2)
    return

  nmillis := long[p_music] * 60 / BPM                           ' adjust timing for bpm
  p_music += 4

  io.freq_out(0, PIEZO, nfreq)                                 ' start the note


con

  ' base timing at 60 BPM

  T_1    = 4000                                                 ' whole note
  T_2    = T_1 >> 1
  T_4    = T_1 >> 2
  T_8    = T_1 >> 3
  T_16   = T_1 >> 4
  T_32   = T_1 >> 5
  T_64   = T_1 >> 6

  T_X    = T_1 >> 7                                             ' to separate notes

  T_D2   = T_2  * 3 / 2                                         ' dotted notes
  T_D4   = T_4  * 3 / 2
  T_D8   = T_8  * 3 / 2
  T_D16  = T_16 * 3 / 2


  ' frequencies in 0.1Hz units

  REST   =    0_0

  N_C0   =   16_3                                              ' natural
  N_Cs0  =   17_3                                              ' sharp
  N_D0   =   18_3
  N_Ds0  =   19_4
  N_E0   =   20_6
  N_F0   =   21_8
  N_Fs0  =   23_1
  N_G0   =   24_5
  N_Gs0  =   25_9
  N_A0   =   27_5
  N_As0  =   29_1
  N_B0   =   30_8

  N_C1   =   32_7
  N_Cs1  =   34_6
  N_D1   =   36_7
  N_Ds1  =   38_8
  N_E1   =   41_2
  N_F1   =   43_6
  N_Fs1  =   46_2
  N_G1   =   49_0
  N_Gs1  =   51_9
  N_A1   =   55_0
  N_As1  =   58_2
  N_B1   =   61_7

  N_C2   =   65_4
  N_Cs2  =   69_3
  N_D2   =   73_4
  N_Ds2  =   77_8
  N_E2   =   82_4
  N_F2   =   87_3
  N_Fs2  =   92_5
  N_G2   =   98_0
  N_Gs2  =  103_8
  N_A2   =  110_0
  N_As2  =  116_5
  N_B2   =  123_5

  N_C3   =  130_8
  N_Cs3  =  138_6
  N_D3   =  146_8
  N_Ds3  =  155_6
  N_E3   =  164_8
  N_F3   =  174_6
  N_Fs3  =  185_0
  N_G3   =  196_0
  N_Gs3  =  207_7
  N_A3   =  220_0
  N_As3  =  233_0
  N_B3   =  247_0

  N_C4   =  261_6
  N_Cs4  =  277_2
  N_D4   =  293_7
  N_Ds4  =  311_1
  N_E4   =  329_6
  N_F4   =  349_2
  N_Fs4  =  370_0
  N_G4   =  392_0
  N_Gs4  =  415_3
  N_A4   =  440_0
  N_As4  =  466_1
  N_B4   =  493_9

  N_C5   =  523_3
  N_Cs5  =  554_4
  N_D5   =  587_3
  N_Ds5  =  622_2
  N_E5   =  659_2
  N_F5   =  698_5
  N_Fs5  =  740_0
  N_G5   =  784_0
  N_Gs5  =  830_6
  N_A5   =  880_0
  N_As5  =  932_3
  N_B5   =  987_8

  N_C6   = 1046_5
  N_Cs6  = 1108_7
  N_D6   = 1174_7
  N_Ds6  = 1244_5
  N_E6   = 1318_5
  N_F6   = 1396_9
  N_Fs6  = 1480_0
  N_G6   = 1568_0
  N_Gs6  = 1661_2
  N_A6   = 1760_0
  N_As6  = 1864_7
  N_B6   = 1975_5

  N_C7   = 2093_0
  N_Cs7  = 2217_5
  N_D7   = 2349_3
  N_Ds7  = 2489_0
  N_E7   = 2637_0
  N_F7   = 2793_8
  N_Fs7  = 2960_0
  N_G7   = 3136_0
  N_Gs7  = 3322_4
  N_A7   = 3520_0
  N_As7  = 3729_3
  N_B7   = 3951_0

  N_C8   = 4186_0
  N_Cs8  = 4434_9
  N_D8   = 4698_6
  N_Ds8  = 4978_0
  N_E8   = 5274_0
  N_F8   = 5587_7
  N_Fs8  = 5919_9
  N_G8   = 6271_9
  N_Gs8  = 6644_9
  N_A8   = 7040_0
  N_As8  = 7458_6
  N_B8   = 7902_1


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}