'' =================================================================================================
''
''   File....... jm_ws2811.spin
''   Purpose.... 800kHz driver for WS2811 pixels
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2016-2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 21 DEC 2024
''
'' =================================================================================================

{ -------------------------------------- }
{  NOTE: Requires system clock >= 80MHz  }
{ -------------------------------------- }

{
  References:
  -- https://cdn-shop.adafruit.com/datasheets/WS2811.pdf


  Important Note:

    In order to accomdate 24- and 32-bit pixels in my drivers, the data for 24-bit pixels
    must be left aligned in long holding the RGB value. In previous drivers, the data was
    right aligned.

    old format: $00_RR_GG_BB
    new format: $RR_GG_BB_00

    BYTE0 of the pixel data is not used in this driver (could be white in other drivers)
}


con

  VERSION = 1_0_0


con { fixed io pins }

  PGM_RX   = 31  { I }                                          ' serial / programming
  PGM_TX   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }


con

  MAX_PIXELS = 1024                                             ' max pixels per strip

  RESET_US   =  500                                             ' 0.5ms reset delay


con { rgbw colors }

  ' borrowed from Gavin Garner's TM1804 LED driver
  ' -- additional colors by Lachlan
  ' -- some alterations by JM
  ' -- modified for RGB and RGBW pixels

  '             RR GG BB WW
  BLACK      = $00_00_00_00
  RED        = $FF_00_00_00
  GREEN      = $00_FF_00_00
  BLUE       = $00_00_FF_00
  WHITE      = $FF_FF_FF_00
  WHITE2     = $00_00_00_FF
  CYAN       = $00_FF_FF_00
  MAGENTA    = $FF_00_FF_00
  YELLOW     = $FF_FF_00_00
  CHARTREUSE = $7F_FF_00_00
  ORANGE     = $FF_60_00_00
  AQUAMARINE = $7F_FF_D4_00
  PINK       = $FF_5F_5F_00
  TURQUOISE  = $3F_E0_C0_00
  REALWHITE  = $C8_FF_FF_00
  INDIGO     = $3F_00_7F_00
  VIOLET     = $BF_7F_BF_00
  MAROON     = $32_00_10_00
  BROWN      = $0E_06_00_00
  CRIMSON    = $DC_28_3C_00
  PURPLE     = $8C_00_FF_00


obj

  gc : "jm_gamma8"                                              ' gamma correction table


var

  long  cog

  long  p_pixels                                                ' pointer to active pixel buffer
  long  npixels                                                 ' number of pixels in buffer
  long  tpin                                                    ' active transmit pin

  ' do not modify order; this structure passed to PASM cog
  '
  long  connection                                              ' compressed connection details
  long  resetticks                                              ' ticks in reset period
  long  t0h                                                     ' bit0 high time (ticks)
  long  t1h                                                     ' bit1 high time (ticks)
  long  cycleticks                                              ' ticks in 1.25us


pub null

'' This is not an application


pub start(p_buf, count, pin) : result | ustix

'' Start smart pixel driver driver
'' -- p_buf is pointer to [long] array holding pixel data
'' -- count is # of pixels supported by array at p_buf
'' -- pin is serial output to pixels

  stop                                                          ' stop if running
  dira[pin] := 0                                                ' clear tx pin in this cog

  if (clkfreq < 80_000_000)                                     ' requires 80MHz clock
    return 0

  ustix := clkfreq / 1_000_000                                  ' ticks in 1us

  ' set cog parameters

  use(p_buf, count, pin)                                        ' set connection details

  resetticks := ustix * RESET_US                                ' note: 80us min reset timing
  t0h        := ustix * 400 / 1000 - 1                          ' pulse widths in ticks (adjusted)
  t1h        := ustix * 800 / 1000 - 1
  cycleticks := ustix * 1_25 / 1_00                             ' ticks in 1.25us (800kHz)

  cog := cognew(@pixdriver, @connection) + 1                    ' start the cog
  if (cog)                                                      ' if it started
    repeat until (connection == 0)                              '  wait until ready
    clear

  return cog


pub stop

'' Stops pixel driver cog (if running)

  if (cog)
    cogstop(cog - 1)
    cog := 0


pub use(p_buf, count, pin) : c

'' Assigns buffer at p_buf to pixel driver
'' -- p_buf is pointer to long array
'' -- count is # of elements in the array
'' -- pin is serial output to pixels
'' -- bits is bit count for pixel type (24 or 32)

   longmove(@p_pixels, @p_buf, 3)                               ' copy parameters
   npixels := 1 #> count <# MAX_PIXELS                          ' force  count into range

   c := p_pixels | ((npixels-1) << 16) | (pin << 26)            ' compress for driver cog

   connection := c                                              ' set new connection


pub connected : result

'' Returns true when latest connection details picked up by driver

  return (connection == 0)


pub color(r, g, b) : rgb

'' Packs r-g-b-w bytes into long

  rgb.byte[3] := r
  rgb.byte[2] := g
  rgb.byte[1] := b


pub colorx(r, g, b, level) : rgb

'' Packs r-g-b-w bytes into long
'' -- r, g, b are element levels, 0..255
'' -- level is brightness, 0..255 (0..100%)

  if (level =< 0)
    rgb := $00_00_00_00

  elseif (level => 255)
    rgb := color(r, g, b)

  else
    r := r * level / 255                                        ' apply level to rgb
    g := g * level / 255
    b := b * level / 255
    rgb := color(r, g, b)


pub wheel(pos) : rgb

'' Creates color from 0 to 255 position input
'' -- colors transition r->g->b back to r

  pos &= $FF

  if (pos < 85)
    return color(255-pos*3, pos*3, 0)
  elseif (pos < 170)
    pos -= 85
    return color(0, 255-pos*3, pos*3)
  else
    pos -= 170
    return color(pos*3, 0, 255-pos*3)


pub wheelx(pos, level) : rgb

'' Creates color from 0 to 255 position input
'' -- colors transition r-g-b back to r
'' -- level is brightness, 0..255 (0..100%)

  pos &= $FF

  if (pos < 85)
    return colorx(255-pos*3, pos*3, 0, level)
  elseif (pos < 170)
    pos -= 85
    return colorx(0, 255-pos*3, pos*3, level)
  else
    pos -= 170
    return colorx(pos*3, 0, 255-pos*3, level)


pub set(ch, rgb)

'' Writes rgb value to channel ch in buffer
'' -- rgbw is packed long in form $RR_GG_BB_00

  if ((ch => 0) and (ch < npixels))
    long[p_pixels][ch] := rgb


pub set_all(rgb)

'' Sets all channels to rgb
'' -- rgb is packed long in form $RR_GG_BB_00

  longfill(p_pixels, rgb, npixels)


pub clear

'' Turns off all LEDs

  longfill(p_pixels, $00_00_00_00, npixels)


pub scale_rgb(rgb, level) : newrgb

'' Scales rgb value to level
'' -- level is brightness, 0..255 (0..100%)

  if (level =< 0)
    newrgb := $00_00_00_00

  elseif (level => 255)
    newrgb := rgb

  else
    newrgb.byte[3] := rgb.byte[3] * level / 255
    newrgb.byte[2] := rgb.byte[2] * level / 255
    newrgb.byte[1] := rgb.byte[1] * level / 255


pub fill(first, last, rgb) | swap

'' Fills first through last channels with rgb
'' -- rgb is packed long in form $RR_GG_BB_00

  first := 0 #> first <# npixels-1
  last  := 0 #> last  <# npixels-1

  if (first > last)
    swap  := first
    first := last
    last  := swap

  longfill(p_pixels+(first << 2), rgb, last-first+1)


pub morph(c1, c2, phase) : newcolor

'' Morph between c1 and c2
'' -- phase 0 is color1
'' -- phase 255 is color2

  if (phase =< 0)
    newcolor := c1

  elseif (phase => 255)
    newcolor := c2

  else
    c1 := scale_rgb(c1, 255-phase)
    c2 := scale_rgb(c2, phase)
    newcolor := c1 + c2


pub read(ch) : rgb

'' Returns color of channel

  if ((ch => 0) and (ch < npixels))                             ' valid?
    return long[p_pixels][ch]
  else
    return $00_00_00_00


pub gamma(idx) : result

'' Adjusts gamma for better midrange colors

  return gc.gamma(0 #> idx <# 255)


pub gamma_rgb(rgb) : result

'' Convert raw rgb to gamma-corrected rgb

  result.byte[3] := gamma(rgb.byte[3])
  result.byte[2] := gamma(rgb.byte[2])
  result.byte[1] := gamma(rgb.byte[1])


pub running : status

'' Returns true if running

  return (cog <> 0)


pub address : p_addr

'' Returns address of assigned pixel array

  return p_pixels


pub num_pixels : count

'' Returns number of pixels in assiged pixel array

  return npixels


dat { auto-run driver }

                        org       0

pixdriver               mov       t1, par                       ' hub address of parameters -> t1
                        movd      :read, #connect               ' location of cog parameters -> :read(dest)
                        mov       t2, #5                        ' get 5 parameters
:read                   rdlong    0-0, t1                       ' copy parameter from hub to cog
                        add       t1, #4                        ' next hub element
                        add       :read, INC_DEST               ' next cog element
                        djnz      t2, #:read                    ' done?

setup                   mov       p_hub, connect                ' extract pointer to pixel array
                        and       p_hub, HX_0000FFFF

                        mov       pixcount, connect             ' extract/fix pixel count
                        shl       pixcount, #6                  ' (remove bits flag, tx pin)
                        shr       pixcount, #22                 ' align, 0..MAX_PIXELS-1
                        add       pixcount, #1                  ' fix, 1..MAX_PIXELS

                        mov       t1, connect                   ' extract pin
                        shl       t1, #1                        ' (remove bits flag)
                        shr       t1, #27                       ' align, 0..31
                        mov       txmask, #1                    ' create mask for tx
                        shl       txmask, t1
                        andn      outa, txmask                  ' set to output low
                        or        dira, txmask

                        mov       t1, #0
                        wrlong    t1, par                         ' tell hub we have connection

rgbx_main               rdlong    connect, par          wz      ' check connection
        if_nz           jmp       #setup

                        mov       addr, p_hub                   ' point to rgbbuf[0]
                        mov       npix, pixcount                ' set # active pixels

frame_loop              rdlong    colorbits, addr               ' read a channel
                        add       addr, #4                      ' point to next

shift_out               mov       nbits, #24                    ' bits for 3 bytes
                        mov       bittimer, cycletix            ' set cycle timer for 800kHz
                        add       bittimer, cnt                 ' sync with system

:loop
                        mov       t1, cnt                       ' mark timing
                        or        outa, txmask                  ' make tx high
                        rcl       colorbits, #1         wc      ' msb --> C
        if_nc           add       t1, bit0hi                    ' if C = 0, use bit0 timing
        if_c            add       t1, bit1hi                    ' if C = 1, use bit1 timing
                        waitcnt   t1, #0                        ' hold high
                        andn      outa, txmask                  ' make tx low
                        waitcnt   bittimer, cycletix            ' finish low side of cycle
                        djnz      nbits, #:loop                 ' more bits?

next_pixel              djnz      npix, #frame_loop             ' done with all leds?

reset_delay             mov       bittimer, resettix            ' set reset timing
                        add       bittimer, cnt                 ' sync timer
                        waitcnt   bittimer, #0                  ' let timer expire

                        jmp       #rgbx_main                    ' back to top

' --------------------------------------------------------------------------------------------------

INC_DEST                long      1 << 9                        ' to increment D field

HX_0000FFFF             long      $0000FFFF                     ' word.0 mask

HX_0000FF00             long      $0000FF00                     ' color byte masks
HX_00FF0000             long      $00FF0000
HX_FF000000             long      $FF000000

connect                 res       1                             ' packed connection details
resettix                res       1                             ' frame reset timing
bit0hi                  res       1                             ' bit0 high timing
bit1hi                  res       1                             ' bit1 high timing
cycletix                res       1                             ' 1.25us cycle ticks

p_hub                   res       1                             ' pointer to pixel buffer in use
pixcount                res       1                             ' # pixels in buffer
txmask                  res       1                             ' mask for output pin

bittimer                res       1                             ' timer for reset/bit
addr                    res       1                             ' address of current rgbw pixel
npix                    res       1                             ' # of pixels to process
colorbits               res       1                             ' rgbw for current pixel
nbits                   res       1                             ' # of bits to process

t1                      res       1                             ' work vars
t2                      res       1
t3                      res       1

                        fit       496


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}